# Generated by Django 2.0.4 on 2018-04-10 11:59

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('spacyal', '0003_auto_20180410_1115'),
    ]

    operations = [
        migrations.AddField(
            model_name='al_project',
            name='users_allowed',
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.0.4 on 2018-04-08 13:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spacyal', '0001_initial'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='case',
            unique_together={('hash', 'suggestion')},
        ),
    ]



# Generated by Django 2.0.4 on 2018-04-10 15:08

from django.db import migrations, models
import spacyal.models


class Migration(migrations.Migration):

    dependencies = [
        ('spacyal', '0004_al_project_users_allowed'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='al_project',
            name='data_path',
        ),
        migrations.AddField(
            model_name='al_project',
            name='model',
            field=models.FileField(blank=True, null=True, upload_to=spacyal.models.get_filepath_model),
        ),
        migrations.AddField(
            model_name='al_project',
            name='texts',
            field=models.FileField(default='', upload_to=spacyal.models.get_filepath_texts),
            preserve_default=False,
        ),
    ]



from rest_framework.views import APIView
from .models import al_project, case
from spacyal.tasks import get_cases, retrain_model
from rest_framework.response import Response
from rest_framework import status
from django.http import HttpResponse
from django_celery_results.models import TaskResult
import json, os
from rest_framework import permissions
from rest_framework.exceptions import PermissionDenied
from shutil import make_archive
from wsgiref.util import FileWrapper
import datetime


class UserALProjectPerm(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        print(obj)
        if request.user.is_superuser:
            return True
        elif request.user in obj.users_allowed.all():
            return True
        raise PermissionDenied({"message": "You don't have permission to"
                                " work on this Project",
                                "object_id": obj.id})


class RetrieveCasesView(APIView):
    """view that allows to retrieve new cases related to a project"""
    permission_classes = (UserALProjectPerm,)

    def get(self, request):
        project_id = request.query_params.get('project_id', None)
        self.check_object_permissions(
            request, al_project.objects.get(pk=project_id))
        t = TaskResult.objects.filter(
            status='SUCCESS',
            result__icontains='"project": {}}}'.format(project_id)).order_by('-id')
        if t.count() > 0:
            c = json.loads(t[0].result)
            get_cases.delay(
                project_id, model=c['folder'], retrained=c['retrained'])
        else:
            get_cases.delay(project_id, retrained=False)
        c = case.objects.filter(project_id=project_id, decission__isnull=True)
        res = [obj.as_dict() for obj in c]
        return Response(res)

    def post(self, request):
        case_id = request.data['case_id']
        decission = request.data['decission']
        retrain = request.data.get('retrain', False)
        project_id = request.data.get('project_id', None)
        c = case.objects.get(pk=case_id)
        c.decission = decission
        c.user = request.user
        c.use = True
        c.save()
        res = {'status': 'saved', 'm_hash': False}
        if retrain:
            t = TaskResult.objects.filter(
                status='SUCCESS',
                result__icontains='"project": {}}}'.format(project_id)).order_by('-id')
            if t.count() > 0:
                c = json.loads(t[0].result)
                m = c['folder'].split('/')[-1]
            else:
                m = 'model_1'
            res['m_hash'] = retrain_model.delay(project_id, m).id
        return Response(res, status=status.HTTP_201_CREATED)


class GetProgressModelView(APIView):

    def get(self, request):
        celery_task = request.query_params.get('celery_task')
        t = TaskResult.objects.get(task_id=celery_task)
        if t.status == 'PROGRESS':
            r = {'status': 'PROGRESS', 'percent': json.loads(t.result)['progress']}
        elif t.status == 'SUCCESS':
            r = {'status': 'SUCCESS', 'percent': 100}
        elif t.status == 'FAILURE':
            r = {'status': 'FAILURE', 'message': t.result}
        return Response(r)


class DownloadModelView(APIView):

    def get(self, request):
        project_id = request.query_params.get('project_id', None)
        project = al_project.objects.get(pk=project_id)
        self.check_object_permissions(
            request, project)
        base_d = '/'.join(project.texts.path.split('/')[:-1])
        t = TaskResult.objects.filter(
            status='SUCCESS',
            result__icontains='"project": {}}}'.format(project_id)).order_by('-id')
        if t.count() > 0:
            c = json.loads(t[0].result)
            model_name = c['folder'].split('/')[-1]
            ts = datetime.datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
            filename = "{}_{}_{}".format(project.name, model_name, ts)
            export_file = make_archive(os.path.join(base_d, filename), 'zip',
                                       base_d, model_name)
            response = HttpResponse(FileWrapper(open(export_file, 'rb')),
                                    content_type='application/zip')
            response['Content-Disposition'] = 'attachment; filename="{}.zip"'.format(
                filename)
            return response


class DownloadCasesView(APIView):

    def get(self, request):
        choices = ('skip', 'correct', 'wrong')
        project_id = request.query_params.get('project_id', None)
        exclude_wrong = request.query_params.get('exclude_wrong', False)
        if exclude_wrong:
            res = al_project.objects.get(pk=project_id).get_training_data(
                include_all=True, include_negative=False)
            for idx, e in enumerate(res):
                res[idx][1]['entities'] = [(x[0], x[1], x[2]) for x in e[1]['entities']]
        else:
            res = al_project.objects.get(pk=project_id).get_training_data(
                include_all=True, include_negative=True)
            for idx, e in enumerate(res):
                res[idx][1]['entities'] = [(x[0], x[1], x[2], choices[x[3]])
                                           for x in e[1]['entities']]
        return Response(res)



# Generated by Django 2.0.4 on 2018-04-08 08:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='al_project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('data_path', models.CharField(help_text='the data folder needs following subfolders: texts, data, model_1, model_2. model_1 should contain the spacy model to start training from.', max_length=255, verbose_name='path of the data folder')),
                ('num_retrain', models.PositiveSmallIntegerField(default=10, verbose_name='number of examples before retraining')),
                ('num_plus_retrain', models.PositiveSmallIntegerField(default=5, verbose_name='number of cases to work ahead')),
                ('ner_type', models.CharField(blank=True, max_length=50, null=True, verbose_name='ner types to work on')),
                ('keep_models', models.BooleanField(default=False, verbose_name='wether to keep old models after retraining')),
                ('al_strategy', models.CharField(choices=[('max score', 'max score'), ('median score', 'median score')], max_length=50, verbose_name='strategy to use for picking cases')),
                ('language_model', models.CharField(default='de', max_length=3, verbose_name='spacy language abbreviation')),
            ],
        ),
        migrations.CreateModel(
            name='case',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('decission', models.PositiveSmallIntegerField(blank=True, choices=[(0, 'skip'), (1, 'correct'), (2, 'wrong')], null=True)),
                ('hash', models.CharField(max_length=255, verbose_name='hash of the sentence file and position')),
                ('sentence', models.TextField(verbose_name='field to store the complete case')),
                ('suggestion', models.CharField(blank=True, max_length=50, null=True, verbose_name='the models suggestion')),
                ('probability', models.FloatField(blank=True, null=True, verbose_name='probability of the model for the case')),
                ('use', models.BooleanField(default=True, verbose_name='whether to use the case in the next iteration')),
                ('start', models.PositiveIntegerField(verbose_name='token offset start')),
                ('end', models.PositiveIntegerField(verbose_name='token offset end')),
                ('start_char', models.PositiveIntegerField(verbose_name='charcter offset start')),
                ('end_char', models.PositiveIntegerField(verbose_name='character offset end')),
                ('text_file', models.CharField(max_length=255, verbose_name='path of the text file')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='spacyal.al_project')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from celery import shared_task, current_task
import random
from pathlib import Path
import spacy
from collections import defaultdict
import os
import pandas as pd
import hashlib
import django
import datetime
import copy
import json
import math
from django.contrib.contenttypes.models import ContentType
from django.db import IntegrityError
#os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'active_learning.settings')
#django.setup()
#from spacyal.models import al_project


def mix_train_data(nlp, TRAIN_DATA):
    res = []
    for text, annotations in TRAIN_DATA:
        doc = nlp(text)
        new_ents = []
        for ner in doc.ents:
            for x in annotations['entities']:
                if x[0] > ner.start_char:
                    new_ents.append((
                        ner.start_char, ner.end_char, ner.label_, 1))
                    break
                elif x[0] == ner.start_char and x[1] == ner.end_char and x[3] == 1:
                    new_ents.append(x)
                    break
                elif x[0] == ner.start_char and x[1] == ner.end_char and x[3] == 2:
                    break
        for ner in annotations:
            if ner not in new_ents and ner[3] == 1:
                new_ents.append(ner)
        res.append((text, {'entities': [(x[0], x[1], x[2]) for x in new_ents]}))
    return res


def dec_use(index, row, lst_cases, lst_ners, lst_hashes):
    z = copy.deepcopy(lst_ners)
    res = False
    for x in lst_cases:
        if x.hash == index:
            if x.decission == 1:
                return False
            elif x.decission == 2 or x.decission == 0:
                try:
                    z.remove(x.suggestion)
                    #print(z, row[z].fillna(0).idxmax())
                    res = row[z].fillna(0).idxmax()
                    print(row[res])
                    if math.isnan(row[res]):
                        res = False
                    elif (index, res) in lst_hashes:
                        res = False
                except Exception as e:
                    print(e)
                    res = False
    return res


@shared_task(time_limit=1800)
def retrain_model(project, model=None, n_iter=50):
    """Load the model, set up the pipeline and train the entity recognizer."""
    if model == 'model_1':
        output_model = 'model_2'
    else:
        output_model = 'model_1'
    al_project = ContentType.objects.get(app_label="spacyal",
                                         model="al_project").model_class()
    project = al_project.objects.get(pk=project)
    base_d = '/'.join(project.texts.path.split('/')[:-1])
    output_dir = os.path.join(base_d, output_model)
    if project.len_training_data() < project.num_retrain:
        message = {'folder': os.path.join(base_d, model),
                   'retrained': False, 'project': project.pk}
        return message
    TRAIN_DATA = project.get_training_data(
        include_all=True, include_negative=True)
    nlp = spacy.load(os.path.join(base_d, model))  # load existing spaCy model
    TRAIN_DATA = mix_train_data(nlp, TRAIN_DATA)
    with open(os.path.join(base_d, 'training_data.json'), 'w') as outp:
        json.dump(TRAIN_DATA, outp)
    r = nlp.get_pipe('ner')
    # get names of other pipes to disable them during training
    other_pipes = [pipe for pipe in nlp.pipe_names if pipe != 'ner']
    count_prog_list = list(range(0, n_iter, int(n_iter/10)))
    with nlp.disable_pipes(*other_pipes):
        optimizer = nlp.begin_training()
        for itn in range(n_iter):
            if itn in count_prog_list:
                current_task.update_state(
                    state='PROGRESS',
                    meta={'progress': count_prog_list.index(itn)*10,
                          'model': output_model, 'project': project.pk})
            random.shuffle(TRAIN_DATA)
            losses = {}
            for text, annotations in TRAIN_DATA:
                nlp.update(
                    [text],  # batch of texts
                    [annotations],  # batch of annotations
                    drop=0.5,  # dropout - make it harder to memorise data
                    sgd=optimizer,  # callable to update weights
                    losses=losses)
    if not Path(output_dir).exists():
        Path(output_dir).mkdir()
    nlp.to_disk(output_dir)
    message = {'folder': output_dir, 'retrained': True, 'project': project.pk}
    return message


@shared_task(time_limit=300)
def get_cases(project, model='model_1', retrained=True):
    al_project = ContentType.objects.get(app_label="spacyal", model="al_project").model_class()
    project = al_project.objects.get(pk=project)
    base_d = '/'.join(project.texts.path.split('/')[:-1])
    case = ContentType.objects.get(app_label="spacyal", model="case").model_class()

    directory = os.fsencode(os.path.join(base_d, 'texts'))
    df_cols = ['start', 'end', 'start_char',
               'end_char', 'sentence', 'dist 1-2', 'max score',
               'median score', 'file']
    df_b = pd.DataFrame(columns=df_cols)
    lst_cases = list(case.objects.filter(project=project).defer(
        'start', 'end', 'start_char', 'end_char', 'text_file', 'user', 'project'
    ))
    lst_hashes = [(x.hash, x.suggestion) for x in lst_cases]
    if retrained or project.last_cases_list is None:
        model = spacy.load(os.path.join(base_d, model))
        for file in os.listdir(directory):
            txt = open(os.path.join(directory, file), 'r').read()
            with model.disable_pipes('ner'):
                doc = model(txt)
            (beams, somethingelse) = model.entity.beam_parse(
                [doc], beam_width=16, beam_density=0.0001)
            entity_scores = defaultdict(float)
            for beam in beams:
                for score, ents in model.entity.moves.get_beam_parses(beam):
                    for start, end, label in ents:
                        hsh = hashlib.md5((os.fsdecode(os.path.join(directory, file))+':'+str(start)+'-'+str(end)).encode('utf8')).hexdigest()
                        entity_scores[(start, end, hsh, label)] += score
            for k in entity_scores.keys():
                if k[0] < 0 or k[1] < 0 or doc[k[0]:k[1]].text == '\n':
                    continue
                df_b.loc[k[2], 'start'] = k[0]
                df_b.loc[k[2], 'end'] = k[1]
                df_b.loc[k[2], 'file'] = os.path.join(directory, file).decode('utf8')
                df_b.loc[k[2], k[3]] = entity_scores[k]
                for sent in doc.sents:
                    if sent.start <= k[0] and sent.end >= k[1]:
                        s = [t.text_with_ws for t in sent if t.i < k[0]]
                        s.append('<mark>'+doc[k[0]:k[1]].text+'</mark>')
                        s.append(doc[k[1]-1].whitespace_)
                        s.extend([t.text_with_ws for t in sent if t.i >= k[1]])
                        df_b.loc[k[2], 'sentence'] = ''.join(s)
                        df_b.loc[k[2], 'start_char'] = doc[:k[0]].end_char
                        df_b.loc[k[2], 'end_char'] = doc[:k[1]].end_char
    elif project.last_cases_list is not None:
        df_b = pd.DataFrame.from_csv(os.path.join(
            base_d, project.last_cases_list))
    lst_ners = [a for a in df_b.columns if a not in df_cols]
    df_b['max score'] = df_b[lst_ners].max(axis=1)
    df_b['median score'] = df_b[lst_ners].median(axis=1)
    df_b = df_b.sort_values(by=[project.al_strategy])
    count = 0
    it_row = df_b.iterrows()
    end_of_examples = False
    while count < (project.num_retrain + project.num_plus_retrain):
        try:
            index, row = next(it_row)
        except StopIteration:
            end_of_examples = True
            break
        s_label = False
        if (index, row[lst_ners].fillna(0).idxmax()) in lst_hashes:
            dec = dec_use(index, row, lst_cases, lst_ners, lst_hashes)
            if dec:
                s_label = dec
            else:
                continue
        else:
            s_label = row[lst_ners].fillna(0).idxmax()
        count += 1
        if s_label and not math.isnan(row[s_label]):
            try:
                c = case.objects.create(
                    sentence=row['sentence'], hash=index, project=project,
                    suggestion=s_label, probability=row[s_label],
                    text_file=row['file'], start=row['start'], end=row['end'], use=True,
                    start_char=row['start_char'], end_char=row['end_char'])
            except IntegrityError:
                print("hash: {} / suggestion: {} already exists".format(index, s_label))
    ts = str(datetime.datetime.now()).split('.')[0].replace(':', '_').replace(' ', '_')
    f_path = os.path.join(
        base_d, 'data/df_cases_{}_{}.csv'.format(project.pk, ts))
    df_b.to_csv(f_path)
    project.last_cases_list = 'data/df_cases_{}_{}.csv'.format(project.pk, ts)
    project.save()
    return 'created {} / {}'.format(f_path, end_of_examples)



from django.urls import path
from . import views

app_name = 'spacyal'

urlpatterns = [
    path('al_project/<int:pk>/', views.AlProject.as_view(),
         name='al_project-detail'),
    path('al_project/list/', views.AlProjectList.as_view(),
         name='al_project-list'),
]



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name='acdh-spacyal',
    version='0.3.1',
    packages=find_packages(
        exclude=['spacyal/__pycache__']),
    include_package_data=True,
    license='MIT License',  # example license
    description='Addon to Django apps that allows to retrain Spacy NER with active learning.',
    long_description=README,
    url='https://github.com/sennierer/spacyal',
    author='Matthias Schlögl',
    author_email='matthias.schloegl@oeaw.ac.at',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',  # replace "X.Y" as appropriate
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',  # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'celery>=4.1.0',
        'amqp==2.2.2',
        'Django>=2.0.4,<3',
        'django-celery-results>=1.0.1',
        'django-crispy-forms>=1.7.2',
        'django-extensions>=2.0.6',
        'djangorestframework>=3.8.0',
        'pandas>=0.22.0',
        'spacy>=2.0.10',
        'django-tables2>=1.21.2'
    ]
)
